import abc
from dataclasses import MISSING, dataclass, fields, is_dataclass
from typing import Any, Generic, TypeAlias, TypeVar

Path: TypeAlias = list[str]

T = TypeVar('T')


@dataclass
class ValueContainer(Generic[T]):
    value: T


class Source(abc.ABC):
    @abc.abstractmethod
    def get(self, path: Path, value_type: type[T]) -> ValueContainer[T] | None:
        ...


class Loader:
    def __init__(self, sources: list[Source]):
        self.sources = sources

    def _load(self, cls: type[T], path: Path = []) -> T | None:
        params: dict[str, Any] = {}
        has_non_default = False
        for field in fields(cls):
            if not field.init:
                continue

            field_path = path + [field.name]

            if is_dataclass(field.type):
                v = self._load(field.type, path=field_path)
                if v is not None:
                    params[field.name] = v
                    has_non_default = True
            else:
                for source in self.sources:
                    v = source.get(path=field_path, value_type=field.type)
                    if v is None:
                        continue

                    params[field.name] = v.value
                    has_non_default = True

            if field.default is not MISSING:
                params[field.name] = field.default

            if field.default_factory is not MISSING:
                params[field.name] = field.default_factory()

        if not has_non_default and path:
            return None

        return cls(**params)

    def load(self, cls: type[T]) -> T:
        value = self._load(cls)
        assert value is not None
        return value



import os
import tempfile
import unittest
from dataclasses import dataclass
from datetime import timedelta

from dcloader import EnvSource, Loader, YAMLSource

timedelta_yaml = """
node: 10s
"""


class TestYaml(unittest.TestCase):
    def test_timedelta(self):
        @dataclass
        class Root:
            node: timedelta

        with tempfile.NamedTemporaryFile() as f:
            f.write(timedelta_yaml.encode("utf8"))
            f.flush()

            loader = Loader([YAMLSource(f.name)])

            obj = loader.load(Root)

            self.assertEqual(obj.node, timedelta(seconds=10))


class TestEnv(unittest.TestCase):
    def setUp(self):
        os.environ["DCLOADER_TEST_NODE"] = "10s"
        os.environ["DCLOADER_TEST_LEAF__NODE"] = "100"

    def test_timedelta(self):
        @dataclass
        class Root:
            node: timedelta

        loader = Loader([EnvSource("DCLOADER_TEST")])

        obj = loader.load(Root)

        self.assertEqual(obj.node, timedelta(seconds=10))

    def test_nested(self):

        @dataclass
        class Leaf:
            node: int

        @dataclass
        class Root:
            leaf: Leaf

        loader = Loader([EnvSource("DCLOADER_TEST")])

        obj = loader.load(Root)

        self.assertEqual(obj.leaf.node, 100)



import os
from datetime import timedelta
from typing import Any, TypeVar

import yaml

from .loader import Path, Source, ValueContainer
from .utils import str_to_timedelta

T = TypeVar("T")


class DictSource(Source):
    def __init__(self, values: dict):
        self.values = values

    def get(self, path: Path, value_type: type[T]) -> ValueContainer[T] | None:
        value: Any = self.values
        for key in path:
            value = value.get(key)
            if value is None:
                return None

        assert isinstance(value, value_type)
        return ValueContainer(value)


class YAMLSource(Source):
    def __init__(self, path: str):
        with open(path, 'r') as file:
            self.values = yaml.safe_load(file)

    def get(self, path: Path, value_type: type[T]) -> ValueContainer[T] | None:
        value = self.values
        for key in path:
            value = value.get(key)
            if value is None:
                return None

        if value_type == timedelta:
            assert type(value) is str
            return ValueContainer(str_to_timedelta(value))  # type: ignore

        assert isinstance(value, value_type)
        return ValueContainer(value)


class EnvSource(Source):
    def __init__(self, prefix: str):
        self.prefix = prefix

    def get(self, path: Path, value_type: type[T]) -> ValueContainer[T] | None:
        name = self.name(path)
        value = os.environ.get(name)
        if value is None:
            return None

        if value_type == timedelta:
            return ValueContainer(str_to_timedelta(value))  # type: ignore

        return ValueContainer(value_type(value))  # type: ignore

    def name(self, path: Path) -> str:
        name = self.prefix + "_"
        name += "__".join(map(lambda x: x.upper(), path))
        return name



from datetime import timedelta


def str_to_timedelta(duration: str) -> timedelta:
    if duration.endswith("s"):
        return timedelta(seconds=int(duration[:-1]))

    if duration.endswith("m"):
        return timedelta(minutes=int(duration[:-1]))

    if duration.endswith("h"):
        return timedelta(hours=int(duration[:-1]))

    if duration.endswith("d"):
        return timedelta(days=int(duration[:-1]))

    raise Exception("not supported")



from distutils.core import setup

setup(
    name="dcloader",
    packages=["dcloader"],
    package_dir={"dcloader": "dcloader"},
    package_data={"dcloader": ["py.typed"]},
    version="0.2.1",
    description="",
    author="Ali Sorour Amini",
    author_email="ali.sorouramini@gmail.com",
    url="https://github.com/alisoam/dcloader",
    install_requires=["PyYAML==6.0.*"],
)



import unittest
from dataclasses import dataclass, field

from dcloader import DictSource, Loader


class Test(unittest.TestCase):
    def test(self):
        @dataclass
        class Root:
            node1: str
            node2: int
            node3: float

        loader = Loader([DictSource({"node1": "value", "node2": 1, "node3": 1.5})])

        obj = loader.load(Root)

        self.assertEqual(obj.node1, "value")
        self.assertEqual(obj.node2, 1)
        self.assertEqual(obj.node3, 1.5)


class TestDataclass(unittest.TestCase):
    def test_init(self):
        @dataclass
        class Root:
            node1: str
            node2: str = field(init=False)

        loader = Loader([DictSource({"node1": "value"})])

        obj = loader.load(Root)

        self.assertEqual(obj.node1, "value")

    def test_default(self):
        @dataclass
        class Root:
            node1: str = "default"
            node2: str = field(default="default")

        loader = Loader([])

        obj = loader.load(Root)

        self.assertEqual(obj.node1, "default")
        self.assertEqual(obj.node2, "default")

    def test_default_factory(self):
        @dataclass
        class Leaf:
            node: str = "default"

        @dataclass
        class Root:
            leaf: Leaf = field(default_factory=lambda: Leaf(node="value"))

        loader = Loader([])

        obj = loader.load(Root)

        self.assertEqual(obj.leaf.node, "value")

    def test_union(self):
        @dataclass
        class Root:
            node1: str | int

        loader1 = Loader([DictSource({"node1": "value"})])
        loader2 = Loader([DictSource({"node1": 1})])

        obj1 = loader1.load(Root)
        obj2 = loader2.load(Root)

        self.assertEqual(obj1.node1, "value")
        self.assertEqual(obj2.node1, 1)
